import pandas as pd
from dagster import DagsterType, dagster_type_loader
from sklearn.base import BaseEstimator

from qdk.utils.typing import is_dataframe, is_mlflow_run, is_model, is_series


@dagster_type_loader(config_schema={})
def load_empty_model(_context, value):  # pylint: disable=unused-argument
    return BaseEstimator()


@dagster_type_loader(config_schema={})
def load_empty_series(_context, value):  # pylint: disable=unused-argument
    return pd.Series()


DataFrameType = DagsterType(
    name="DataFrameType",
    type_check_fn=lambda _, value: is_dataframe(value),
    description="Can represent either a pandas or dask dataframe.",
)
SeriesType = DagsterType(
    name="SeriesType",
    type_check_fn=lambda _, value: is_series(value),
    description="Can represent either a pandas or dask series.",
    loader=load_empty_series,
)
ModelType = DagsterType(
    name="ModelType",
    type_check_fn=lambda _, value: is_model(value),
    description="Generic typing for machine learning models.",
    loader=load_empty_model,
)
MLFlowRunType = DagsterType(
    name="MLFlowRunType",
    type_check_fn=lambda _, value: is_mlflow_run(value),
    description="A type that represents a qdk.MLFlowRun class.",
)



from typing import Any, Dict, Union

import dask.dataframe as dd
import pandas as pd
from dagster import Field, OutputDefinition, Permissive
from qdk.base import BaseComponent
from qdk.dagster_types import DataFrameType
from qdk.s3_connection import S3Connection


class DataFrameLoader(BaseComponent):
    compute_function = "load"
    tags = {
        "kind": "loader",
    }
    input_defs = []
    output_defs = [
        OutputDefinition(DataFrameType, "df"),
    ]
    config_schema = {
        "uri": Field(
            str,
            description="The uri to load the dataframe from.",
        ),
        "use_dask": Field(
            bool,
            default_value=False,
            description="Whether to load the dataframe using Dask.",
        ),
        "repartitions": Field(
            int,
            is_required=False,
            description="How many partitions to create.",
        ),
        "drop_na": Field(
            bool,
            default_value=False,
            description="Whether to drop rows with missing values.",
        ),
        "load_params": Field(
            Permissive({}),
            description="Extra parameters that get passed to the loading function.",
        ),
    }

    @classmethod
    def load(
        cls,
        uri: str,
        use_dask: bool = False,
        repartitions: int = None,
        drop_na: bool = False,
        load_params: Dict[str, Any] = {},
    ) -> Union[pd.DataFrame, dd.DataFrame]:
        # Choose which framework to use for loading the data
        framework = dd if use_dask else pd

        # Inject S3 connection information into the load parameters
        # if the uri starts with an S3 connection indicator
        if uri.startswith("s3://"):
            s3_connection = S3Connection()
            load_params = {
                "storage_options": {
                    "client_kwargs": {
                        "aws_access_key_id": s3_connection.aws_access_key_id,
                        "aws_secret_access_key": s3_connection.aws_access_secret_key,
                        "endpoint_url": s3_connection.aws_endpoint_url,
                    }
                },
                **load_params,
            }

        if uri.endswith(".csv"):
            df = framework.read_csv(
                uri,
                **load_params,
            )

        if uri.endswith(".parquet"):
            df = framework.read_parquet(
                uri,
                **load_params,
            )

        elif (
            uri.endswith(".json")
            or uri.endswith(".jsonl")
            or uri.endswith(".jsonlines")
        ):
            df = framework.read_json(
                uri,
                orient="records",
                lines=True,
                **load_params,
            )

        elif uri.endswith(".pkl"):
            df = pd.read_pickle(
                uri,
                **load_params,
            )

            if use_dask:
                df = dd.from_pandas(
                    df,
                    npartitions=repartitions,
                )

        # If using dask and repartitions are supplied
        if use_dask and repartitions is not None:
            df = df.repartition(npartitions=repartitions)

        # Drop rows with missing values
        if drop_na:
            df = df.dropna()
            df = df.reset_index(drop=True)

        return df



from typing import Any, Dict, List, Sequence, Set

from dagster import (
    Field,
    InputDefinition,
    OutputDefinition,
    SolidDefinition,
    SolidExecutionContext,
)
from dagster.core.definitions.events import Output


class BaseComponent:
    input_defs: List[InputDefinition] = []
    output_defs: List[OutputDefinition] = []
    config_schema: Dict[str, Field] = {}
    required_resource_keys: Set[str] = set()
    tags: Dict[str, str] = {}

    @classmethod
    def _compute_hook(cls, step_context: SolidExecutionContext, inputs: Dict[str, Any]):
        """The compute hook that runs the actual compute function.
            Needs to overridden by the descendant classes.

        Args:
            step_context (SolidExecutionContext): The Dagster execution context.
            inputs (Dict[str, Any]): The inputs that are supplied to the solid as a dictionary.

        Raises:
            NotImplementedError: The compute hook is not implemented by default.
        """
        # Run the compute function ("cls.compute_function" needs to be defined by the children)
        # The inputs of the previous and solid config are injected into the compute function
        results = getattr(cls, cls.compute_function)(
            **inputs, **step_context.solid_config
        )

        # If the results are not a sequence, wrap it in a list
        if not isinstance(results, Sequence):
            results = [results]

        # Zip the results with the output definitions and yield the results
        for result, output_def in zip(results, cls.output_defs):
            yield Output(result, output_def.name)

    @classmethod
    def to_solid(cls, name: str) -> SolidDefinition:
        """Generates a solid definition for this component.

        Args:
            name (str): The name of the solid.
            input_defs (List[InputDefinition], optional): The list of Dagster input values. Defaults to [].
            output_defs (List[OutputDefinition], optional): The list of Dagster output values. Defaults to [].

        Returns:
            SolidDefinition: The Solid that is used by Dagster to execute this component .
        """
        return SolidDefinition(
            name=name,
            input_defs=cls.input_defs,
            output_defs=cls.output_defs,
            required_resource_keys=cls.required_resource_keys,
            config_schema=cls.config_schema,
            compute_fn=cls._compute_hook,
            tags=cls.tags,
        )



from typing import Union
from pandas.core.series import Series
from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd
import dask.dataframe as dd


class ArrayTransformer(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        return self

    def transform(
        self,
        X: Union[pd.DataFrame, dd.DataFrame],
        y: Union[pd.Series, dd.Series] = None,
    ):

        if isinstance(X, pd.DataFrame):
            return X.to_numpy()
        elif isinstance(X, dd.DataFrame):
            return X.to_dask_array()



import pandas as pd
import dask.dataframe as dd
from dagster import InputDefinition
from qdk.dagster_types import DataFrameType, SeriesType
from qdk.training.base import TrainingComponent
from typing import Union
from sklearn.base import BaseEstimator


class SklearnComponent(TrainingComponent):
    input_defs = [
        InputDefinition("X", DataFrameType),
        InputDefinition("y", SeriesType),
        InputDefinition("model", BaseEstimator),
    ]
    required_resource_keys = {"mlflow"}

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        y: Union[pd.Series, dd.Series],
        model: BaseEstimator,
    ):
        return model.fit(X, y)



from .dagster_types import DataFrameType, MLFlowRunType, ModelType, SeriesType
from .inference.base import InferenceComponent
from .inference.mlflow import MLFlowInferenceComponent
from .loader.dataframe import DataFrameLoader
from .resources.io_manager import io_manager
from .s3_connection import S3Connection
from .training.base import TrainingComponent
from .training.doc2vec import Doc2VecTrainingComponent
from .training.grid_search import GridSearchTrainingComponent
from .training.mlflow import MLFlowTrainingComponent
from .training.sklearn import SklearnComponent
from .training.word2vec import Word2VecTrainingComponent
from .transform.array import ArrayTransformer
from .transform.base import TransformComponent
from .transform.noun_phrase import NounPhraseComponent
from .transform.tokenize import TokenizeComponent
from .transform.train_test import TrainTestComponent
from .transform.yake import YakeComponent



from typing import Dict

import pandas as pd
from dagster import Any, InputDefinition, Output, OutputDefinition
from qdk.base import BaseComponent


class InferenceComponent(BaseComponent):
    compute_function = "predict"
    tags = {
        "kind": "inference",
    }
    input_defs = [
        InputDefinition("data", pd.DataFrame),
        InputDefinition("model", Any),
    ]
    output_defs = [
        OutputDefinition(pd.DataFrame, "predictions"),
    ]

    @classmethod
    def predict(cls, data: Dict[str, pd.DataFrame], model: Any):
        raise NotImplementedError(
            'Make sure you added a "predict" function to the component'
        )
